/*******************************************************************************
Copyright (c) 2010, Jonathan Hiller (Cornell University)
If used in publication cite "J. Hiller and H. Lipson "Dynamic Simulation of Soft Heterogeneous Objects" In press. (2011)"

This file is part of Voxelyze.
Voxelyze is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
Voxelyze is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.
See <http://www.opensource.org/licenses/lgpl-3.0.html> for license details.
*******************************************************************************/

#ifndef VXS_VOXEL_H
#define VXS_VOXEL_H

#include "VXt_Voxel.h"
#include "Quat3D.h"


class CVXS_BondCollision;

//References:
//http://gafferongames.com/game-physics/physics-in-3d/

//Poissons ratio (volumetric expansion)
//http://www.colorado.edu/engineering/CAS/courses.d/Structures.d/IAST.Lect05.d/IAST.Lect05.pdf


//info about a voxel that changes during a simulation (i.e. state)
class CVXS_Voxel : public CVXt_Voxel
{
public:
	CVXS_Voxel(CVX_Sim* pSimIn, int SIndexIn, int XIndexIn, int MatIndexIn, Vec3D<>& NominalPositionIn, vfloat OriginalScaleIn);
	~CVXS_Voxel(void);
	CVXS_Voxel(const CVXS_Voxel& VIn) : CVXt_Voxel(VIn) {*this = VIn;} //copy constructor
	CVXS_Voxel& operator=(const CVXS_Voxel& VIn);

	void ResetVoxel(); //resets this voxel to its default (imported) state.

	void EulerStep(); //updates the state of the voxel based on the current forces and moments.

	//Collisions
	bool LinkColBond(int CBondIndex); //collision bond index...
	void UpdateColBondPointers(); //updates all links (pointers) to bonds according top current p_Sim
	void UnlinkColBonds();

	//input information
	inline void SetInputForce(const Vec3D<>& InputForceIn) {InputForce = InputForceIn;} //adds a specified force to this voxel. Subsequent calls over-write this force. (Can be used in conjunction with picking in gui to drag voxels around)
	inline void ScaleExternalInputs(const vfloat ScaleFactor=1.0) {ExternalInputScale=ScaleFactor;} //scales force, torque, etc. to some percentage of its set value

	//Get info about the current state of this voxel
	const inline Vec3D<> GetCurPos() const {return Pos;}
	const inline Vec3D<double> GetCurPosHighAccuracy(void) const {return Pos;}
	const inline Quat3D<> GetCurAngle() const {return Angle;}
	const inline Quat3D<double> GetCurAngleHighAccuracy(void) const {return Angle;}
	const inline float GetCurScale() const {return Scale;}
	const inline Vec3D<> GetCurVel() const {return Vel;}
	const inline Vec3D<> GetCurAngVel() const {return AngVel;}
	const inline vfloat GetPressure() const {return Pressure;}
	const inline vfloat GetCurKineticE() const {return KineticEnergy;}
	const inline bool GetCurStaticFric() const {return StaticFricFlag;}
	const inline vfloat GetCurAbsDisp() const {return (Vec3D<>(Pos)-GetNominalPosition()).Length();}
	inline Vec3D<> GetSizeCurrent() const {return CornerPosCur-CornerNegCur;}
	inline Vec3D<> GetCornerPos() const {return CornerPosCur;}
	inline Vec3D<> GetCornerNeg() const {return CornerNegCur;}

	vfloat GetMaxBondStrain() const;
	vfloat GetMaxBondStrainE() const;
	vfloat GetMaxBondStress() const;
	vfloat GetCurGroundPenetration(); //how far into the ground penetrating (penetration is positive, no penetration is zero)
	Vec3D<> GetCurForce(bool forceRecalc = false) {if (forceRecalc) CalcTotalForce(); return ForceCurrent;} //just returns last calculated force

	//yeilded and broken flags
	inline void SetYielded(const bool Yielded) {VYielded = Yielded;}
	inline bool GetYielded() const {return VYielded;}
	inline void SetBroken(const bool Broken) {VBroken = Broken;}
	inline bool GetBroken() const {return VBroken;}

	//utilities
	inline void ZeroMotion() {LinMom = Vec3D<double>(0,0,0); AngMom = Vec3D<double>(0,0,0); Vel = Vec3D<>(0,0,0); AngVel = Vec3D<>(0,0,0); KineticEnergy = 0;}
	vfloat CalcVoxMatStress(const vfloat StrainIn, bool* const IsPastYielded, bool* const IsPastFail) const;

	//display color stuff
	void SetColor(float r, float g, float b, float a);

	//Poissons!
//	void SetStrainDir(BondDir Bond, vfloat StrainIn);
//	Vec3D<> StrainPosDirsCur, StrainNegDirsCur; //cache the strain in each bond direction
//	vfloat GetVoxelStrain(Axis DesiredAxis);


	Vec3D<> poissonsstrain;

private:
	//State variable of this voxel being simulated
	Vec3D<double> Pos; //translation
	Vec3D<double> LinMom;
	Quat3D<double> Angle; //rotation
	Vec3D<double> AngMom;
	vfloat Scale; //nominal scale based on temperature, etc.
	//Vec3D<double> Scale; //nominal scale based on temperature, poissons deformation, etc.

	Vec3D<> CornerPosCur, CornerNegCur; //actual size based on volume effects or other deviations from Scale
	bool StaticFricFlag; //flag to set if this voxel shouldnot move in X/Y due to being in static friction regime
	bool VYielded, VBroken; //is this voxel yielded or brokem according to the current material model?
		
	//convenience derived secondary state quantities:
	Vec3D<> Vel;
	vfloat KineticEnergy;
	Vec3D<> AngVel;
	vfloat Pressure;


	//current input parameters
	vfloat ExternalInputScale; //Scales the external Force, displacement, torque (range: 0-1)
	Vec3D<> InputForce; //current input force (general purpose force acting on this voxel)

	//collision system information
	std::vector<int> ColBondInds; //collision bond indices
	std::vector<CVXS_BondCollision*> ColBondPointers; //collision bond pointers
	inline bool IAmVox2Col(const int BondDirColIndex) const; //returns true if this voxel is Vox2 of the specified bond

	//current display color of this voxel.
	float m_Red, m_Green, m_Blue, m_Trans; //can update voxel color based on state, mode, etc.

	//force calculations of this voxel
	Vec3D<> ForceCurrent; //cached current force, as last calculated by CalcTotalForce()

	Vec3D<> CalcTotalForce(); //calculates the total force acting on this voxel (without fixed constraints...)
	Vec3D<> CalcTotalMoment(); //Calculates the total moment action on this voxel

	//secondary force calculations
	Vec3D<> CalcFloorEffect(Vec3D<> TotalVoxForce); //calculates the object's interaction with a floor under gravity
	Vec3D<> CalcGndDampEffect(); //damps everything to ground as quick as possible...


	//new
	Vec3D<> strain(bool tensionStrain); //LCS returns voxel strain. if tensionStrain true and no actual tension in that
	void updatePoissonsstrain();

};

#endif //VXS_VOXEL_H
